#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import List, Optional
from pyspark.errors import PySparkTypeError


def validate_optional_list_of_str_arg(arg_name: str, arg_value: Optional[List[str]]) -> None:
    """
    Validates types for Optional[list] parameters.

    Validation succeeds if both:
    - The value provided to the parameter is either a list or None.
    - All elements inside the list are instances of the expected element type.

    Otherwise, raises an error.
    """
    if arg_value is None:
        return

    if not isinstance(arg_value, list):
        raise PySparkTypeError(
            errorClass="NOT_LIST_OF_STR",
            messageParameters={
                "arg_name": arg_name,
                "arg_type": type(arg_value).__name__,
            },
        )

    for el in arg_value:
        if not isinstance(el, str):
            raise PySparkTypeError(
                errorClass="NOT_LIST_OF_STR",
                messageParameters={
                    "arg_name": arg_name,
                    "arg_type": type(arg_value).__name__,
                },
            )



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
import functools
import inspect
import itertools
import os
import threading
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Match,
    TypeVar,
    Type,
    Optional,
    Union,
    overload,
    cast,
)
from types import FrameType

import pyspark
from pyspark.errors.error_classes import ERROR_CLASSES_MAP

T = TypeVar("T")
FuncT = TypeVar("FuncT", bound=Callable[..., Any])

_current_origin = threading.local()

# Providing DataFrame debugging options to reduce performance slowdown.
# Default is True.
_enable_debugging_cache = None


def is_debugging_enabled() -> bool:
    global _enable_debugging_cache

    if _enable_debugging_cache is None:
        from pyspark.sql import SparkSession

        spark = SparkSession.getActiveSession()
        if spark is not None:
            _enable_debugging_cache = (
                spark.conf.get(
                    "spark.python.sql.dataFrameDebugging.enabled",
                    "true",  # type: ignore[union-attr]
                ).lower()
                == "true"
            )
        else:
            _enable_debugging_cache = False

    return _enable_debugging_cache


def current_origin() -> threading.local:
    global _current_origin

    if not hasattr(_current_origin, "fragment"):
        _current_origin.fragment = None
    if not hasattr(_current_origin, "call_site"):
        _current_origin.call_site = None
    return _current_origin


def set_current_origin(fragment: Optional[str], call_site: Optional[str]) -> None:
    global _current_origin

    _current_origin.fragment = fragment
    _current_origin.call_site = call_site


class ErrorClassesReader:
    """
    A reader to load error information from error-conditions.json.
    """

    def __init__(self) -> None:
        self.error_info_map = ERROR_CLASSES_MAP

    def get_error_message(self, errorClass: str, messageParameters: Dict[str, str]) -> str:
        """
        Returns the completed error message by applying message parameters to the message template.
        """
        message_template = self.get_message_template(errorClass)
        # Verify message parameters.
        message_parameters_from_template = re.findall("<([a-zA-Z0-9_-]+)>", message_template)
        assert set(message_parameters_from_template) == set(messageParameters), (
            f"Undefined error message parameter for error class: {errorClass}. "
            f"Parameters: {messageParameters}"
        )

        def replace_match(match: Match[str]) -> str:
            return match.group().translate(str.maketrans("<>", "{}"))

        # Convert <> to {} only when paired.
        message_template = re.sub(r"<([^<>]*)>", replace_match, message_template)

        return message_template.format(**messageParameters)

    def get_message_template(self, errorClass: str) -> str:
        """
        Returns the message template for corresponding error class from error-conditions.json.

        For example,
        when given `errorClass` is "EXAMPLE_ERROR_CLASS",
        and corresponding error class in error-conditions.json looks like the below:

        .. code-block:: python

            "EXAMPLE_ERROR_CLASS" : {
              "message" : [
                "Problem <A> because of <B>."
              ]
            }

        In this case, this function returns:
        "Problem <A> because of <B>."

        For sub error class, when given `errorClass` is "EXAMPLE_ERROR_CLASS.SUB_ERROR_CLASS",
        and corresponding error class in error-conditions.json looks like the below:

        .. code-block:: python

            "EXAMPLE_ERROR_CLASS" : {
              "message" : [
                "Problem <A> because of <B>."
              ],
              "sub_class" : {
                "SUB_ERROR_CLASS" : {
                  "message" : [
                    "Do <C> to fix the problem."
                  ]
                }
              }
            }

        In this case, this function returns:
        "Problem <A> because <B>. Do <C> to fix the problem."
        """
        error_classes = errorClass.split(".")
        len_error_classes = len(error_classes)
        assert len_error_classes in (1, 2)

        # Generate message template for main error class.
        main_error_class = error_classes[0]
        if main_error_class in self.error_info_map:
            main_error_class_info_map = self.error_info_map[main_error_class]
        else:
            raise ValueError(f"Cannot find main error class '{main_error_class}'")

        main_message_template = "\n".join(main_error_class_info_map["message"])
        if "breaking_change_info" in main_error_class_info_map:
            main_message_template += " " + "\n".join(
                main_error_class_info_map["breaking_change_info"]["migration_message"]
            )

        has_sub_class = len_error_classes == 2

        if not has_sub_class:
            message_template = main_message_template
        else:
            # Generate message template for sub error class if exists.
            sub_error_class = error_classes[1]
            main_error_class_subclass_info_map = main_error_class_info_map["sub_class"]
            if sub_error_class in main_error_class_subclass_info_map:
                sub_error_class_info_map = main_error_class_subclass_info_map[sub_error_class]
            else:
                raise ValueError(f"Cannot find sub error class '{sub_error_class}'")

            sub_message_template = "\n".join(sub_error_class_info_map["message"])
            if "breaking_change_info" in sub_error_class_info_map:
                sub_message_template += " " + "\n".join(
                    sub_error_class_info_map["breaking_change_info"]["migration_message"]
                )
            message_template = main_message_template + " " + sub_message_template

        return message_template

    def get_breaking_change_info(self, errorClass: Optional[str]) -> Optional[Dict[str, Any]]:
        """
        Returns the breaking change info for an error if it is present.
        """
        if errorClass is None:
            return None
        error_classes = errorClass.split(".")
        len_error_classes = len(error_classes)
        assert len_error_classes in (1, 2)

        main_error_class = error_classes[0]
        if main_error_class in self.error_info_map:
            main_error_class_info_map = self.error_info_map[main_error_class]
        else:
            raise ValueError(f"Cannot find main error class '{main_error_class}'")

        if len_error_classes == 2:
            sub_error_class = error_classes[1]
            main_error_class_subclass_info_map = main_error_class_info_map["sub_class"]
            if sub_error_class in main_error_class_subclass_info_map:
                sub_error_class_info_map = main_error_class_subclass_info_map[sub_error_class]
            else:
                raise ValueError(f"Cannot find sub error class '{sub_error_class}'")
            if "breaking_change_info" in sub_error_class_info_map:
                return sub_error_class_info_map["breaking_change_info"]
        if "breaking_change_info" in main_error_class_info_map:
            return main_error_class_info_map["breaking_change_info"]
        return None


def _capture_call_site(depth: int) -> str:
    """
    Capture the call site information including file name, line number, and function name.
    This function updates the thread-local storage from JVM side (PySparkCurrentOrigin)
    with the current call site information when a PySpark API function is called.

    Notes
    -----
    The call site information is used to enhance error messages with the exact location
    in the user code that led to the error.
    """
    # Filtering out PySpark code and keeping user code only
    pyspark_root = os.path.dirname(pyspark.__file__)

    def inspect_stack() -> Iterator[FrameType]:
        frame = inspect.currentframe()
        while frame:
            yield frame
            frame = frame.f_back

    stack = (f for f in inspect_stack() if pyspark_root not in f.f_code.co_filename)

    selected_frames: Iterator[FrameType] = itertools.islice(stack, depth)

    # We try import here since IPython is not a required dependency
    try:
        import IPython

        # ipykernel is required for IPython
        import ipykernel  # type: ignore[import-not-found]

        ipython = IPython.get_ipython()
        # Filtering out IPython related frames
        ipy_root = os.path.dirname(IPython.__file__)
        ipykernel_root = os.path.dirname(ipykernel.__file__)
        selected_frames = (
            frame
            for frame in selected_frames
            if (ipy_root not in frame.f_code.co_filename)
            and (ipykernel_root not in frame.f_code.co_filename)
        )
    except ImportError:
        ipython = None

    # Identifying the cell is useful when the error is generated from IPython Notebook
    if ipython:
        call_sites = [
            f"line {frame.f_lineno} in cell [{ipython.execution_count}]"
            for frame in selected_frames
        ]
    else:
        call_sites = [f"{frame.f_code.co_filename}:{frame.f_lineno}" for frame in selected_frames]
    call_sites_str = "\n".join(call_sites)

    return call_sites_str


def _with_origin(func: FuncT) -> FuncT:
    """
    A decorator to capture and provide the call site information to the server side
    when PySpark API functions are invoked.
    """

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        from pyspark.sql import SparkSession
        from pyspark.sql.utils import is_remote

        if hasattr(func, "__name__") and is_debugging_enabled():
            if is_remote():
                # Getting the configuration requires RPC call. Uses the default value for now.
                depth = 1
                set_current_origin(func.__name__, _capture_call_site(depth))

                try:
                    return func(*args, **kwargs)
                finally:
                    set_current_origin(None, None)
            else:
                spark = SparkSession.getActiveSession()
                if spark is None:
                    return func(*args, **kwargs)
                assert spark._jvm is not None
                jvm_pyspark_origin = getattr(
                    spark._jvm, "org.apache.spark.sql.catalyst.trees.PySparkCurrentOrigin"
                )
                depth = int(
                    spark.conf.get(  # type: ignore[arg-type]
                        "spark.sql.stackTracesInDataFrameContext"
                    )
                )
                # Update call site when the function is called
                jvm_pyspark_origin.set(func.__name__, _capture_call_site(depth))

                try:
                    return func(*args, **kwargs)
                finally:
                    jvm_pyspark_origin.clear()
        else:
            return func(*args, **kwargs)

    return cast(FuncT, wrapper)


@overload
def with_origin_to_class(cls_or_ignores: Type[T], ignores: Optional[List[str]] = None) -> Type[T]:
    ...


@overload
def with_origin_to_class(
    cls_or_ignores: Optional[List[str]] = None,
) -> Callable[[Type[T]], Type[T]]:
    ...


def with_origin_to_class(
    cls_or_ignores: Optional[Union[Type[T], List[str]]] = None, ignores: Optional[List[str]] = None
) -> Union[Type[T], Callable[[Type[T]], Type[T]]]:
    """
    Decorate all methods of a class with `_with_origin` to capture call site information.
    """
    if cls_or_ignores is None or isinstance(cls_or_ignores, list):
        ignores = cls_or_ignores or []
        return lambda cls: with_origin_to_class(cls, ignores)
    else:
        cls = cls_or_ignores
        if os.environ.get("PYSPARK_PIN_THREAD", "true").lower() == "true":
            skipping = set(
                ["__init__", "__new__", "__iter__", "__nonzero__", "__repr__", "__bool__"]
                + (ignores or [])
            )
            for name, method in cls.__dict__.items():
                # Excluding Python magic methods that do not utilize JVM functions.
                if callable(method) and name not in skipping:
                    setattr(cls, name, _with_origin(method))
        return cls



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Any


__all__ = ["StreamingListener"]


class StreamingListener:
    def __init__(self) -> None:
        pass

    def onStreamingStarted(self, streamingStarted: Any) -> None:
        """
        Called when the streaming has been started.
        """
        pass

    def onReceiverStarted(self, receiverStarted: Any) -> None:
        """
        Called when a receiver has been started
        """
        pass

    def onReceiverError(self, receiverError: Any) -> None:
        """
        Called when a receiver has reported an error
        """
        pass

    def onReceiverStopped(self, receiverStopped: Any) -> None:
        """
        Called when a receiver has been stopped
        """
        pass

    def onBatchSubmitted(self, batchSubmitted: Any) -> None:
        """
        Called when a batch of jobs has been submitted for processing.
        """
        pass

    def onBatchStarted(self, batchSubmitted: Any) -> None:
        """
        Called when processing of a batch of jobs has started.
        """
        pass

    def onBatchCompleted(self, batchCompleted: Any) -> None:
        """
        Called when processing of a batch of jobs has completed.
        """
        pass

    def onOutputOperationStarted(self, outputOperationStarted: Any) -> None:
        """
        Called when processing of a job of a batch has started.
        """
        pass

    def onOutputOperationCompleted(self, outputOperationCompleted: Any) -> None:
        """
        Called when processing of a job of a batch has completed
        """
        pass

    class Java:
        implements = ["org.apache.spark.streaming.api.java.PythonStreamingListener"]



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import struct
import sys
import unittest
import difflib
import functools
from decimal import Decimal
from time import time, sleep
import signal
from typing import (
    Any,
    Optional,
    Union,
    Dict,
    List,
    Callable,
)
from itertools import zip_longest

from pyspark import SparkConf
from pyspark.errors import PySparkAssertionError, PySparkException, PySparkTypeError
from pyspark.errors.exceptions.base import QueryContextType
from pyspark.sql.dataframe import DataFrame
from pyspark.sql import Row
from pyspark.sql.types import StructType, StructField, VariantVal
from pyspark.sql.functions import col, when


__all__ = ["assertDataFrameEqual", "assertSchemaEqual"]


def have_package(name: str) -> bool:
    import importlib

    return importlib.util.find_spec(name) is not None


have_numpy = have_package("numpy")
numpy_requirement_message = None if have_numpy else "No module named 'numpy'"

have_scipy = have_package("scipy")
scipy_requirement_message = None if have_scipy else "No module named 'scipy'"

have_sklearn = have_package("sklearn")
sklearn_requirement_message = None if have_sklearn else "No module named 'sklearn'"

have_torch = have_package("torch")
torch_requirement_message = None if have_torch else "No module named 'torch'"

have_torcheval = have_package("torcheval")
torcheval_requirement_message = None if have_torcheval else "No module named 'torcheval'"

have_deepspeed = have_package("deepspeed")
deepspeed_requirement_message = None if have_deepspeed else "No module named 'deepspeed'"

have_plotly = have_package("plotly")
plotly_requirement_message = None if have_plotly else "No module named 'plotly'"

have_matplotlib = have_package("matplotlib")
matplotlib_requirement_message = None if have_matplotlib else "No module named 'matplotlib'"

have_tabulate = have_package("tabulate")
tabulate_requirement_message = None if have_tabulate else "No module named 'tabulate'"

have_graphviz = have_package("graphviz")
graphviz_requirement_message = None if have_graphviz else "No module named 'graphviz'"

have_flameprof = have_package("flameprof")
flameprof_requirement_message = None if have_flameprof else "No module named 'flameprof'"

have_jinja2 = have_package("jinja2")
jinja2_requirement_message = None if have_jinja2 else "No module named 'jinja2'"

have_openpyxl = have_package("openpyxl")
openpyxl_requirement_message = None if have_openpyxl else "No module named 'openpyxl'"

have_yaml = have_package("yaml")
yaml_requirement_message = None if have_yaml else "No module named 'yaml'"

have_grpc = have_package("grpc")
grpc_requirement_message = None if have_grpc else "No module named 'grpc'"

have_grpc_status = have_package("grpc_status")
grpc_status_requirement_message = None if have_grpc_status else "No module named 'grpc_status'"


googleapis_common_protos_requirement_message = None

try:
    from google.rpc import error_details_pb2
except ImportError as e:
    googleapis_common_protos_requirement_message = str(e)
have_googleapis_common_protos = googleapis_common_protos_requirement_message is None

pandas_requirement_message = None
try:
    from pyspark.sql.pandas.utils import require_minimum_pandas_version

    require_minimum_pandas_version()
except Exception as e:
    # If Pandas version requirement is not satisfied, skip related tests.
    pandas_requirement_message = str(e)

have_pandas = pandas_requirement_message is None


pyarrow_requirement_message = None
try:
    from pyspark.sql.pandas.utils import require_minimum_pyarrow_version

    require_minimum_pyarrow_version()
except Exception as e:
    # If Arrow version requirement is not satisfied, skip related tests.
    pyarrow_requirement_message = str(e)

have_pyarrow = pyarrow_requirement_message is None


connect_requirement_message = (
    pandas_requirement_message
    or pyarrow_requirement_message
    or grpc_requirement_message
    or googleapis_common_protos_requirement_message
    or grpc_status_requirement_message
)

should_test_connect = connect_requirement_message is None


is_ansi_mode_test = True
if os.environ.get("SPARK_ANSI_SQL_MODE") == "false":
    is_ansi_mode_test = False

ansi_mode_not_supported_message = "ANSI mode is not supported" if is_ansi_mode_test else None


def read_int(b):
    return struct.unpack("!i", b)[0]


def write_int(i):
    return struct.pack("!i", i)


def timeout(timeout):
    def decorator(func):
        def handler(signum, frame):
            raise TimeoutError(f"Function {func.__name__} timed out after {timeout} seconds")

        def wrapper(*args, **kwargs):
            signal.alarm(0)
            signal.signal(signal.SIGALRM, handler)
            signal.alarm(timeout)
            try:
                result = func(*args, **kwargs)
            finally:
                signal.alarm(0)
            return result

        return wrapper

    return decorator


def eventually(
    timeout=30.0,
    catch_assertions=False,
    catch_timeout=False,
):
    """
    Wait a given amount of time for a condition to pass, else fail with an error.
    This is a helper utility for PySpark tests.

    Parameters
    ----------
    condition : function
        Function that checks for termination conditions. condition() can return:
            - True or None: Conditions met. Return without error.
            - other value: Conditions not met yet. Continue. Upon timeout,
              include last such value in error message.
              Note that this method may be called at any time during
              streaming execution (e.g., even before any results
              have been created).
    timeout : int
        Number of seconds to wait.  Default 30 seconds.
    catch_assertions : bool
        If False (default), do not catch AssertionErrors.
        If True, catch AssertionErrors; continue, but save
        error to throw upon timeout.
    catch_timeout : bool
        If False (default), do not catch TimeoutError.
        If True, catch TimeoutError; continue, but save
        error to throw upon timeout.
    """
    assert timeout > 0
    assert isinstance(catch_assertions, bool)
    assert isinstance(catch_timeout, bool)

    def decorator(condition: Callable) -> Callable:
        assert isinstance(condition, Callable)

        @functools.wraps(condition)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            start_time = time()
            lastValue = None
            numTries = 0
            while time() - start_time < timeout:
                numTries += 1

                try:
                    lastValue = condition(*args, **kwargs)
                except AssertionError as e:
                    if catch_assertions:
                        lastValue = e
                    else:
                        raise e
                except TimeoutError as e:
                    if catch_timeout:
                        lastValue = e
                    else:
                        raise e

                if lastValue is True or lastValue is None:
                    return

                print(f"\nAttempt #{numTries} failed!\n{lastValue}")
                sleep(0.01)

            if isinstance(lastValue, (AssertionError, TimeoutError)):
                raise lastValue
            else:
                raise AssertionError(
                    "Test failed due to timeout after %g sec, with last condition returning: %s"
                    % (timeout, lastValue)
                )

        return wrapper

    return decorator


class QuietTest:
    def __init__(self, sc):
        self.log4j = sc._jvm.org.apache.log4j

    def __enter__(self):
        self.old_level = self.log4j.LogManager.getRootLogger().getLevel()
        self.log4j.LogManager.getRootLogger().setLevel(self.log4j.Level.FATAL)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.log4j.LogManager.getRootLogger().setLevel(self.old_level)


class PySparkTestCase(unittest.TestCase):
    def setUp(self):
        from pyspark import SparkContext

        self._old_sys_path = list(sys.path)
        class_name = self.__class__.__name__
        self.sc = SparkContext("local[4]", class_name)

    def tearDown(self):
        self.sc.stop()
        sys.path = self._old_sys_path


class ReusedPySparkTestCase(unittest.TestCase):
    @classmethod
    def conf(cls):
        """
        Override this in subclasses to supply a more specific conf
        """
        return SparkConf()

    @classmethod
    def setUpClass(cls):
        from pyspark import SparkContext

        cls.sc = SparkContext(cls.master(), cls.__name__, conf=cls.conf())

    @classmethod
    def master(cls):
        return "local[4]"

    @classmethod
    def tearDownClass(cls):
        cls.sc.stop()

    def test_assert_classic_mode(self):
        from pyspark.sql import is_remote

        self.assertFalse(is_remote())

    def quiet(self):
        from pyspark.testing.utils import QuietTest

        return QuietTest(self.sc)


class ByteArrayOutput:
    def __init__(self):
        self.buffer = bytearray()

    def write(self, b):
        self.buffer += b

    def close(self):
        pass


def _terminal_color_support():
    try:
        # determine if environment supports color
        script = "$(test $(tput colors)) && $(test $(tput colors) -ge 8) && echo true || echo false"
        return os.popen(script).read()
    except Exception:
        return False


def _context_diff(actual: List[str], expected: List[str], n: int = 3):
    """
    Modified from difflib context_diff API,
    see original code here: https://github.com/python/cpython/blob/main/Lib/difflib.py#L1180
    """

    def red(s: str) -> str:
        red_color = "\033[31m"
        no_color = "\033[0m"
        return red_color + str(s) + no_color

    prefix = dict(insert="+ ", delete="- ", replace="! ", equal="  ")
    for group in difflib.SequenceMatcher(None, actual, expected).get_grouped_opcodes(n):
        yield "*** actual ***"
        if any(tag in {"replace", "delete"} for tag, _, _, _, _ in group):
            for tag, i1, i2, _, _ in group:
                for line in actual[i1:i2]:
                    if tag != "equal" and _terminal_color_support():
                        yield red(prefix[tag] + str(line))
                    else:
                        yield prefix[tag] + str(line)

        yield "\n"

        yield "*** expected ***"
        if any(tag in {"replace", "insert"} for tag, _, _, _, _ in group):
            for tag, _, _, j1, j2 in group:
                for line in expected[j1:j2]:
                    if tag != "equal" and _terminal_color_support():
                        yield red(prefix[tag] + str(line))
                    else:
                        yield prefix[tag] + str(line)


class PySparkErrorTestUtils:
    """
    This util provide functions to accurate and consistent error testing
    based on PySpark error classes.
    """

    def check_error(
        self,
        exception: PySparkException,
        errorClass: str,
        messageParameters: Optional[Dict[str, str]] = None,
        query_context_type: Optional[QueryContextType] = None,
        fragment: Optional[str] = None,
        matchPVals: bool = False,
    ):
        query_context = exception.getQueryContext()
        assert bool(query_context) == (query_context_type is not None), (
            "`query_context_type` is required when QueryContext exists. "
            f"QueryContext: {query_context}."
        )
        # Test if given error is an instance of PySparkException.
        self.assertIsInstance(
            exception,
            PySparkException,
            f"checkError requires 'PySparkException', got '{exception.__class__.__name__}'.",
        )

        # Test error class
        expected = errorClass
        actual = exception.getCondition()
        self.assertEqual(
            expected, actual, f"Expected error class was '{expected}', got '{actual}'."
        )

        # Test message parameters
        expected = messageParameters
        actual = exception.getMessageParameters()
        if matchPVals:
            self.assertEqual(
                len(expected),
                len(actual),
                "Expected message parameters count does not match actual message parameters count"
                f": {len(expected)}, {len(actual)}.",
            )
            for key, value in expected.items():
                self.assertIn(
                    key,
                    actual,
                    f"Expected message parameter key '{key}' was not found "
                    "in actual message parameters.",
                )
                self.assertRegex(
                    actual[key],
                    value,
                    f"Expected message parameter value '{value}' does not match actual message "
                    f"parameter value '{actual[key]}'.",
                ),
        else:
            self.assertEqual(
                expected, actual, f"Expected message parameters was '{expected}', got '{actual}'"
            )

        # Test query context
        if query_context:
            expected = query_context_type
            actual_contexts = exception.getQueryContext()
            for actual_context in actual_contexts:
                actual = actual_context.contextType()
                self.assertEqual(
                    expected, actual, f"Expected QueryContext was '{expected}', got '{actual}'"
                )
                if actual == QueryContextType.DataFrame:
                    assert (
                        fragment is not None
                    ), "`fragment` is required when QueryContextType is DataFrame."
                    expected = fragment
                    actual = actual_context.fragment()
                    self.assertEqual(
                        expected,
                        actual,
                        f"Expected PySpark fragment was '{expected}', got '{actual}'",
                    )


def assertSchemaEqual(
    actual: StructType,
    expected: StructType,
    ignoreNullable: bool = True,
    ignoreColumnOrder: bool = False,
    ignoreColumnName: bool = False,
):
    __tracebackhide__ = True
    r"""
    A util function to assert equality between DataFrame schemas `actual` and `expected`.

    .. versionadded:: 3.5.0

    Parameters
    ----------
    actual : StructType
        The DataFrame schema that is being compared or tested.
    expected : StructType
        The expected schema, for comparison with the actual schema.
    ignoreNullable : bool, default True
        Specifies whether a column’s nullable property is included when checking for
        schema equality.
        When set to `True` (default), the nullable property of the columns being compared
        is not taken into account and the columns will be considered equal even if they have
        different nullable settings.
        When set to `False`, columns are considered equal only if they have the same nullable
        setting.
        .. versionadded:: 4.0.0
    ignoreColumnOrder : bool, default False
        Specifies whether to compare columns in the order they appear in the DataFrame or by
        column name.
        If set to `False` (default), columns are compared in the order they appear in the
        DataFrames.
        When set to `True`, a column in the expected DataFrame is compared to the column with the
        same name in the actual DataFrame.
        .. versionadded:: 4.0.0
    ignoreColumnName : bool, default False
        Specifies whether to fail the initial schema equality check if the column names in the two
        DataFrames are different.
        When set to `False` (default), column names are checked and the function fails if they are
        different.
        When set to `True`, the function will succeed even if column names are different.
        Column data types are compared for columns in the order they appear in the DataFrames.
        .. versionadded:: 4.0.0

    Notes
    -----
    When assertSchemaEqual fails, the error message uses the Python `difflib` library to display
    a diff log of the `actual` and `expected` schemas.

    Examples
    --------
    >>> from pyspark.sql.types import StructType, StructField, ArrayType, IntegerType, DoubleType
    >>> s1 = StructType([StructField("names", ArrayType(DoubleType(), True), True)])
    >>> s2 = StructType([StructField("names", ArrayType(DoubleType(), True), True)])
    >>> assertSchemaEqual(s1, s2)  # pass, schemas are identical

    Different schemas with `ignoreNullable=False` would fail.

    >>> s3 = StructType([StructField("names", ArrayType(DoubleType(), True), False)])
    >>> assertSchemaEqual(s1, s3, ignoreNullable=False)  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    PySparkAssertionError: [DIFFERENT_SCHEMA] Schemas do not match.
    --- actual
    +++ expected
    - StructType([StructField('names', ArrayType(DoubleType(), True), True)])
    ?                                                                 ^^^
    + StructType([StructField('names', ArrayType(DoubleType(), True), False)])
    ?                                                                 ^^^^


    >>> df1 = spark.createDataFrame(data=[(1, 1000), (2, 3000)], schema=["id", "number"])
    >>> df2 = spark.createDataFrame(data=[("1", 1000), ("2", 5000)], schema=["id", "amount"])
    >>> assertSchemaEqual(df1.schema, df2.schema)  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    PySparkAssertionError: [DIFFERENT_SCHEMA] Schemas do not match.
    --- actual
    +++ expected
    - StructType([StructField('id', LongType(), True), StructField('number', LongType(), True)])
    ?                               ^^                               ^^^^^
    + StructType([StructField('id', StringType(), True), StructField('amount', LongType(), True)])
    ?                               ^^^^                              ++++ ^

    Compare two schemas ignoring the column order.

    >>> s1 = StructType(
    ...     [StructField("a", IntegerType(), True), StructField("b", DoubleType(), True)]
    ... )
    >>> s2 = StructType(
    ...     [StructField("b", DoubleType(), True), StructField("a", IntegerType(), True)]
    ... )
    >>> assertSchemaEqual(s1, s2, ignoreColumnOrder=True)

    Compare two schemas ignoring the column names.

    >>> s1 = StructType(
    ...     [StructField("a", IntegerType(), True), StructField("c", DoubleType(), True)]
    ... )
    >>> s2 = StructType(
    ...     [StructField("b", IntegerType(), True), StructField("d", DoubleType(), True)]
    ... )
    >>> assertSchemaEqual(s1, s2, ignoreColumnName=True)
    """
    if not isinstance(actual, StructType):
        raise PySparkTypeError(
            errorClass="NOT_STRUCT",
            messageParameters={"arg_name": "actual", "arg_type": type(actual).__name__},
        )
    if not isinstance(expected, StructType):
        raise PySparkTypeError(
            errorClass="NOT_STRUCT",
            messageParameters={"arg_name": "expected", "arg_type": type(expected).__name__},
        )

    def compare_schemas_ignore_nullable(s1: StructType, s2: StructType):
        if len(s1) != len(s2):
            return False
        zipped = zip_longest(s1, s2)
        for sf1, sf2 in zipped:
            if not compare_structfields_ignore_nullable(sf1, sf2):
                return False
        return True

    def compare_structfields_ignore_nullable(actualSF: StructField, expectedSF: StructField):
        if actualSF is None and expectedSF is None:
            return True
        elif actualSF is None or expectedSF is None:
            return False
        if actualSF.name != expectedSF.name:
            return False
        else:
            return compare_datatypes_ignore_nullable(actualSF.dataType, expectedSF.dataType)

    def compare_datatypes_ignore_nullable(dt1: Any, dt2: Any):
        # checks datatype equality, using recursion to ignore nullable
        if dt1.typeName() == dt2.typeName():
            if dt1.typeName() == "array":
                return compare_datatypes_ignore_nullable(dt1.elementType, dt2.elementType)
            elif dt1.typeName() == "decimal":
                # Fix for SPARK-51062: Compare precision and scale for decimal types
                return dt1.precision == dt2.precision and dt1.scale == dt2.scale
            elif dt1.typeName() == "struct":
                return compare_schemas_ignore_nullable(dt1, dt2)
            else:
                return True
        else:
            return False

    if ignoreColumnOrder:
        actual = StructType(sorted(actual, key=lambda x: x.name))
        expected = StructType(sorted(expected, key=lambda x: x.name))

    if ignoreColumnName:
        actual = StructType(
            [StructField(str(i), field.dataType, field.nullable) for i, field in enumerate(actual)]
        )
        expected = StructType(
            [
                StructField(str(i), field.dataType, field.nullable)
                for i, field in enumerate(expected)
            ]
        )

    if (ignoreNullable and not compare_schemas_ignore_nullable(actual, expected)) or (
        not ignoreNullable and actual != expected
    ):
        generated_diff = difflib.ndiff(str(actual).splitlines(), str(expected).splitlines())
        error_msg = "\n".join(generated_diff)
        raise PySparkAssertionError(
            errorClass="DIFFERENT_SCHEMA",
            messageParameters={"error_msg": error_msg},
        )


from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import pandas
    import pyspark.pandas


def assertDataFrameEqual(
    actual: Union[DataFrame, "pandas.DataFrame", "pyspark.pandas.DataFrame", List[Row]],
    expected: Union[DataFrame, "pandas.DataFrame", "pyspark.pandas.DataFrame", List[Row]],
    checkRowOrder: bool = False,
    rtol: float = 1e-5,
    atol: float = 1e-8,
    ignoreNullable: bool = True,
    ignoreColumnOrder: bool = False,
    ignoreColumnName: bool = False,
    ignoreColumnType: bool = False,
    maxErrors: Optional[int] = None,
    showOnlyDiff: bool = False,
    includeDiffRows=False,
):
    __tracebackhide__ = True
    r"""
    A util function to assert equality between `actual` and `expected`
    (DataFrames or lists of Rows), with optional parameters `checkRowOrder`, `rtol`, and `atol`.

    Supports Spark, Spark Connect, pandas, and pandas-on-Spark DataFrames.
    For more information about pandas-on-Spark DataFrame equality, see the docs for
    `assertPandasOnSparkEqual`.

    .. versionadded:: 3.5.0

    Parameters
    ----------
    actual : DataFrame (Spark, Spark Connect, pandas, or pandas-on-Spark) or list of Rows
        The DataFrame that is being compared or tested.
    expected : DataFrame (Spark, Spark Connect, pandas, or pandas-on-Spark) or list of Rows
        The expected result of the operation, for comparison with the actual result.
    checkRowOrder : bool, optional
        A flag indicating whether the order of rows should be considered in the comparison.
        If set to `False` (default), the row order is not taken into account.
        If set to `True`, the order of rows is important and will be checked during comparison.
        (See Notes)
    rtol : float, optional
        The relative tolerance, used in asserting approximate equality for float values in actual
        and expected. Set to 1e-5 by default. (See Notes)
    atol : float, optional
        The absolute tolerance, used in asserting approximate equality for float values in actual
        and expected. Set to 1e-8 by default. (See Notes)
    ignoreNullable : bool, default True
        Specifies whether a column’s nullable property is included when checking for
        schema equality.
        When set to `True` (default), the nullable property of the columns being compared
        is not taken into account and the columns will be considered equal even if they have
        different nullable settings.
        When set to `False`, columns are considered equal only if they have the same nullable
        setting.

        .. versionadded:: 4.0.0
    ignoreColumnOrder : bool, default False
        Specifies whether to compare columns in the order they appear in the DataFrame or by
        column name.
        If set to `False` (default), columns are compared in the order they appear in the
        DataFrames.
        When set to `True`, a column in the expected DataFrame is compared to the column with the
        same name in the actual DataFrame.

        .. versionadded:: 4.0.0
    ignoreColumnName : bool, default False
        Specifies whether to fail the initial schema equality check if the column names in the two
        DataFrames are different.
        When set to `False` (default), column names are checked and the function fails if they are
        different.
        When set to `True`, the function will succeed even if column names are different.
        Column data types are compared for columns in the order they appear in the DataFrames.

        .. versionadded:: 4.0.0
    ignoreColumnType : bool, default False
        Specifies whether to ignore the data type of the columns when comparing.
        When set to `False` (default), column data types are checked and the function fails if they
        are different.
        When set to `True`, the schema equality check will succeed even if column data types are
        different and the function will attempt to compare rows.

        .. versionadded:: 4.0.0
    maxErrors : bool, optional
        The maximum number of row comparison failures to encounter before returning.
        When this number of row comparisons have failed, the function returns independent of how
        many rows have been compared.
        Set to None by default which means compare all rows independent of number of failures.

        .. versionadded:: 4.0.0
    showOnlyDiff : bool, default False
        If set to `True`, the error message will only include rows that are different.
        If set to `False` (default), the error message will include all rows
        (when there is at least one row that is different).

        .. versionadded:: 4.0.0
    includeDiffRows: bool, False
        If set to `True`, the unequal rows are included in PySparkAssertionError for further
        debugging. If set to `False` (default), the unequal rows are not returned as a data set.

        .. versionadded:: 4.0.0

    Notes
    -----
    When `assertDataFrameEqual` fails, the error message uses the Python `difflib` library to
    display a diff log of each row that differs in `actual` and `expected`.

    For `checkRowOrder`, note that PySpark DataFrame ordering is non-deterministic, unless
    explicitly sorted.

    Note that schema equality is checked only when `expected` is a DataFrame (not a list of Rows).

    For DataFrames with float/decimal values, assertDataFrame asserts approximate equality.
    Two float/decimal values a and b are approximately equal if the following equation is True:

    ``absolute(a - b) <= (atol + rtol * absolute(b))``.

    `ignoreColumnOrder` cannot be set to `True` if `ignoreColumnNames` is also set to `True`.
    `ignoreColumnNames` cannot be set to `True` if `ignoreColumnOrder` is also set to `True`.

    Examples
    --------
    >>> df1 = spark.createDataFrame(data=[("1", 1000), ("2", 3000)], schema=["id", "amount"])
    >>> df2 = spark.createDataFrame(data=[("1", 1000), ("2", 3000)], schema=["id", "amount"])
    >>> assertDataFrameEqual(df1, df2)  # pass, DataFrames are identical

    >>> df1 = spark.createDataFrame(data=[("1", 0.1), ("2", 3.23)], schema=["id", "amount"])
    >>> df2 = spark.createDataFrame(data=[("1", 0.109), ("2", 3.23)], schema=["id", "amount"])
    >>> assertDataFrameEqual(df1, df2, rtol=1e-1)  # pass, DataFrames are approx equal by rtol

    >>> df1 = spark.createDataFrame(data=[(1, 1000), (2, 3000)], schema=["id", "amount"])
    >>> list_of_rows = [Row(1, 1000), Row(2, 3000)]
    >>> assertDataFrameEqual(df1, list_of_rows)  # pass, actual and expected data are equal

    >>> import pyspark.pandas as ps  # doctest: +SKIP
    >>> df1 = ps.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6], 'c': [7, 8, 9]})  # doctest: +SKIP
    >>> df2 = ps.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6], 'c': [7, 8, 9]})  # doctest: +SKIP
    >>> # pass, pandas-on-Spark DataFrames are equal
    >>> assertDataFrameEqual(df1, df2)  # doctest: +SKIP

    >>> df1 = spark.createDataFrame(
    ...     data=[("1", 1000.00), ("2", 3000.00), ("3", 2000.00)], schema=["id", "amount"])
    >>> df2 = spark.createDataFrame(
    ...     data=[("1", 1001.00), ("2", 3000.00), ("3", 2003.00)], schema=["id", "amount"])
    >>> assertDataFrameEqual(df1, df2)  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    PySparkAssertionError: [DIFFERENT_ROWS] Results do not match: ( 66.66667 % )
    *** actual ***
    ! Row(id='1', amount=1000.0)
      Row(id='2', amount=3000.0)
    ! Row(id='3', amount=2000.0)
    *** expected ***
    ! Row(id='1', amount=1001.0)
      Row(id='2', amount=3000.0)
    ! Row(id='3', amount=2003.0)

    Example for ignoreNullable

    >>> from pyspark.sql.types import StructType, StructField, StringType, LongType
    >>> df1_nullable = spark.createDataFrame(
    ...     data=[(1000, "1"), (5000, "2")],
    ...     schema=StructType(
    ...         [StructField("amount", LongType(), True), StructField("id", StringType(), True)]
    ...     )
    ... )
    >>> df2_nullable = spark.createDataFrame(
    ...     data=[(1000, "1"), (5000, "2")],
    ...     schema=StructType(
    ...         [StructField("amount", LongType(), True), StructField("id", StringType(), False)]
    ...     )
    ... )
    >>> assertDataFrameEqual(df1_nullable, df2_nullable, ignoreNullable=True)  # pass
    >>> assertDataFrameEqual(
    ...     df1_nullable, df2_nullable, ignoreNullable=False
    ... )  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    PySparkAssertionError: [DIFFERENT_SCHEMA] Schemas do not match.
    --- actual
    +++ expected
    - StructType([StructField('amount', LongType(), True), StructField('id', StringType(), True)])
    ?                                                                                      ^^^
    + StructType([StructField('amount', LongType(), True), StructField('id', StringType(), False)])
    ?                                                                                      ^^^^

    Example for ignoreColumnOrder

    >>> df1_col_order = spark.createDataFrame(
    ...     data=[(1000, "1"), (5000, "2")], schema=["amount", "id"]
    ... )
    >>> df2_col_order = spark.createDataFrame(
    ...     data=[("1", 1000), ("2", 5000)], schema=["id", "amount"]
    ... )
    >>> assertDataFrameEqual(df1_col_order, df2_col_order, ignoreColumnOrder=True)

    Example for ignoreColumnName

    >>> df1_col_names = spark.createDataFrame(
    ...     data=[(1000, "1"), (5000, "2")], schema=["amount", "identity"]
    ... )
    >>> df2_col_names = spark.createDataFrame(
    ...     data=[(1000, "1"), (5000, "2")], schema=["amount", "id"]
    ... )
    >>> assertDataFrameEqual(df1_col_names, df2_col_names, ignoreColumnName=True)

    Example for ignoreColumnType

    >>> df1_col_types = spark.createDataFrame(
    ...     data=[(1000, "1"), (5000, "2")], schema=["amount", "id"]
    ... )
    >>> df2_col_types = spark.createDataFrame(
    ...     data=[(1000.0, "1"), (5000.0, "2")], schema=["amount", "id"]
    ... )
    >>> assertDataFrameEqual(df1_col_types, df2_col_types, ignoreColumnType=True)

    Example for maxErrors (will only report the first mismatching row)

    >>> df1 = spark.createDataFrame([(1, "A"), (2, "B"), (3, "C")])
    >>> df2 = spark.createDataFrame([(1, "A"), (2, "X"), (3, "Y")])
    >>> assertDataFrameEqual(df1, df2, maxErrors=1)  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    PySparkAssertionError: [DIFFERENT_ROWS] Results do not match: ( 33.33333 % )
    *** actual ***
      Row(_1=1, _2='A')
    ! Row(_1=2, _2='B')
    *** expected ***
      Row(_1=1, _2='A')
    ! Row(_1=2, _2='X')

    Example for showOnlyDiff (will only report the mismatching rows)

    >>> df1 = spark.createDataFrame([(1, "A"), (2, "B"), (3, "C")])
    >>> df2 = spark.createDataFrame([(1, "A"), (2, "X"), (3, "Y")])
    >>> assertDataFrameEqual(df1, df2, showOnlyDiff=True)  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    PySparkAssertionError: [DIFFERENT_ROWS] Results do not match: ( 66.66667 % )
    *** actual ***
    ! Row(_1=2, _2='B')
    ! Row(_1=3, _2='C')
    *** expected ***
    ! Row(_1=2, _2='X')
    ! Row(_1=3, _2='Y')

    The `includeDiffRows` parameter can be used to include the rows that did not match
    in the PySparkAssertionError. This can be useful for debugging or further analysis.

    >>> df1 = spark.createDataFrame(
    ...     data=[("1", 1000.00), ("2", 3000.00), ("3", 2000.00)], schema=["id", "amount"])
    >>> df2 = spark.createDataFrame(
    ...     data=[("1", 1001.00), ("2", 3000.00), ("3", 2003.00)], schema=["id", "amount"])
    >>> try:
    ...     assertDataFrameEqual(df1, df2, includeDiffRows=True)
    ... except PySparkAssertionError as e:
    ...     spark.createDataFrame(e.data).show()  # doctest: +NORMALIZE_WHITESPACE
    +-----------+-----------+
    |         _1|         _2|
    +-----------+-----------+
    |{1, 1000.0}|{1, 1001.0}|
    |{3, 2000.0}|{3, 2003.0}|
    +-----------+-----------+
    """
    if actual is None and expected is None:
        return True
    elif actual is None:
        raise PySparkAssertionError(
            errorClass="INVALID_TYPE_DF_EQUALITY_ARG",
            messageParameters={
                "expected_type": "Union[DataFrame, ps.DataFrame, List[Row]]",
                "arg_name": "actual",
                "actual_type": None,
            },
        )
    elif expected is None:
        raise PySparkAssertionError(
            errorClass="INVALID_TYPE_DF_EQUALITY_ARG",
            messageParameters={
                "expected_type": "Union[DataFrame, ps.DataFrame, List[Row]]",
                "arg_name": "expected",
                "actual_type": None,
            },
        )

    has_pandas = False
    try:
        # If pandas dependencies are available, allow pandas or pandas-on-Spark DataFrame
        import pandas as pd

        has_pandas = True
    except ImportError:
        # no pandas, so we won't call pandasutils functions
        pass

    has_arrow = False
    try:
        import pyarrow

        has_arrow = True
    except ImportError:
        pass

    if has_pandas and has_arrow:
        import pyspark.pandas as ps
        from pyspark.testing.pandasutils import PandasOnSparkTestUtils

        if (
            isinstance(actual, pd.DataFrame)
            or isinstance(expected, pd.DataFrame)
            or isinstance(actual, ps.DataFrame)
            or isinstance(expected, ps.DataFrame)
        ):
            # handle pandas DataFrames
            # assert approximate equality for float data
            return PandasOnSparkTestUtils().assert_eq(
                actual, expected, almost=True, rtol=rtol, atol=atol, check_row_order=checkRowOrder
            )

    if not isinstance(actual, (DataFrame, list)):
        raise PySparkAssertionError(
            errorClass="INVALID_TYPE_DF_EQUALITY_ARG",
            messageParameters={
                "expected_type": "Union[DataFrame, ps.DataFrame, List[Row]]",
                "arg_name": "actual",
                "actual_type": type(actual),
            },
        )
    elif not isinstance(expected, (DataFrame, list)):
        raise PySparkAssertionError(
            errorClass="INVALID_TYPE_DF_EQUALITY_ARG",
            messageParameters={
                "expected_type": "Union[DataFrame, ps.DataFrame, List[Row]]",
                "arg_name": "expected",
                "actual_type": type(expected),
            },
        )

    if ignoreColumnOrder:
        actual = actual.select(*sorted(actual.columns))
        expected = expected.select(*sorted(expected.columns))

    def rename_dataframe_columns(df: DataFrame) -> DataFrame:
        """Rename DataFrame columns to sequential numbers for comparison"""
        renamed_columns = [str(i) for i in range(len(df.columns))]
        return df.toDF(*renamed_columns)

    if ignoreColumnName:
        actual = rename_dataframe_columns(actual)
        expected = rename_dataframe_columns(expected)

    def cast_columns_to_string(df: DataFrame) -> DataFrame:
        """Cast all DataFrame columns to string for comparison"""
        for col_name in df.columns:
            # Add logic to remove trailing .0 for float columns that are whole numbers
            df = df.withColumn(
                col_name,
                when(
                    (col(col_name).cast("float").isNotNull())
                    & (col(col_name).cast("float") == col(col_name).cast("int")),
                    col(col_name).cast("int").cast("string"),
                ).otherwise(col(col_name).cast("string")),
            )
        return df

    if ignoreColumnType:
        actual = cast_columns_to_string(actual)
        expected = cast_columns_to_string(expected)

    def compare_rows(r1: Row, r2: Row):
        def compare_vals(val1, val2):
            if isinstance(val1, list) and isinstance(val2, list):
                return len(val1) == len(val2) and all(
                    compare_vals(x, y) for x, y in zip(val1, val2)
                )
            elif isinstance(val1, Row) and isinstance(val2, Row):
                return all(compare_vals(x, y) for x, y in zip(val1, val2))
            elif isinstance(val1, dict) and isinstance(val2, dict):
                return (
                    len(val1.keys()) == len(val2.keys())
                    and val1.keys() == val2.keys()
                    and all(compare_vals(val1[k], val2[k]) for k in val1.keys())
                )
            elif isinstance(val1, float) and isinstance(val2, float):
                if abs(val1 - val2) > (atol + rtol * abs(val2)):
                    return False
            elif isinstance(val1, Decimal) and isinstance(val2, Decimal):
                if abs(val1 - val2) > (Decimal(atol) + Decimal(rtol) * abs(val2)):
                    return False
            elif isinstance(val1, VariantVal) and isinstance(val2, VariantVal):
                return compare_vals(val1.toPython(), val2.toPython())
            else:
                if val1 != val2:
                    return False
            return True

        if r1 is None and r2 is None:
            return True
        elif r1 is None or r2 is None:
            return False

        return compare_vals(r1, r2)

    def assert_rows_equal(
        rows1: List[Row], rows2: List[Row], maxErrors: int = None, showOnlyDiff: bool = False
    ):
        __tracebackhide__ = True
        zipped = list(zip_longest(rows1, rows2))
        diff_rows_cnt = 0
        diff_rows = []
        has_diff_rows = False

        rows_str1 = ""
        rows_str2 = ""

        # count different rows
        for r1, r2 in zipped:
            if not compare_rows(r1, r2):
                diff_rows_cnt += 1
                has_diff_rows = True
                if includeDiffRows:
                    diff_rows.append((r1, r2))
                rows_str1 += str(r1) + "\n"
                rows_str2 += str(r2) + "\n"
                if maxErrors is not None and diff_rows_cnt >= maxErrors:
                    break
            elif not showOnlyDiff:
                rows_str1 += str(r1) + "\n"
                rows_str2 += str(r2) + "\n"

        generated_diff = _context_diff(
            actual=rows_str1.splitlines(), expected=rows_str2.splitlines(), n=len(zipped)
        )

        if has_diff_rows:
            error_msg = "Results do not match: "
            percent_diff = (diff_rows_cnt / len(zipped)) * 100
            error_msg += "( %.5f %% )" % percent_diff
            error_msg += "\n" + "\n".join(generated_diff)
            data = diff_rows if includeDiffRows else None
            raise PySparkAssertionError(
                errorClass="DIFFERENT_ROWS", messageParameters={"error_msg": error_msg}, data=data
            )

    # only compare schema if expected is not a List
    if not isinstance(actual, list) and not isinstance(expected, list):
        if ignoreNullable:
            assertSchemaEqual(actual.schema, expected.schema)
        elif actual.schema != expected.schema:
            generated_diff = difflib.ndiff(
                str(actual.schema).splitlines(), str(expected.schema).splitlines()
            )
            error_msg = "\n".join(generated_diff)

            raise PySparkAssertionError(
                errorClass="DIFFERENT_SCHEMA",
                messageParameters={"error_msg": error_msg},
            )

    if not isinstance(actual, list):
        if actual.isStreaming:
            raise PySparkAssertionError(
                errorClass="UNSUPPORTED_OPERATION",
                messageParameters={"operation": "assertDataFrameEqual on streaming DataFrame"},
            )
        actual_list = actual.collect()
    else:
        actual_list = actual

    if not isinstance(expected, list):
        if expected.isStreaming:
            raise PySparkAssertionError(
                errorClass="UNSUPPORTED_OPERATION",
                messageParameters={"operation": "assertDataFrameEqual on streaming DataFrame"},
            )
        expected_list = expected.collect()
    else:
        expected_list = expected

    if not checkRowOrder:
        # rename duplicate columns for sorting
        actual_list = sorted(actual_list, key=lambda x: str(x))
        expected_list = sorted(expected_list, key=lambda x: str(x))

    assert_rows_equal(actual_list, expected_list, maxErrors=maxErrors, showOnlyDiff=showOnlyDiff)


def _test() -> None:
    import doctest
    from pyspark.sql import SparkSession
    import pyspark.testing.utils

    globs = pyspark.testing.utils.__dict__.copy()
    spark = SparkSession.builder.master("local[4]").appName("testing.utils tests").getOrCreate()
    globs["spark"] = spark
    (failure_count, test_count) = doctest.testmod(
        pyspark.testing.utils,
        globs=globs,
        optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE,
    )
    spark.stop()
    if failure_count:
        sys.exit(-1)


if __name__ == "__main__":
    _test()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Implementation of spark-pipelines CLI.

Example usage:
    $ bin/spark-pipelines run --spec /path/to/pipeline.yaml
"""
from contextlib import contextmanager
import argparse
import glob
import importlib.util
import os
import yaml
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Generator, List, Mapping, Optional, Sequence

from pyspark.errors import PySparkException, PySparkTypeError
from pyspark.sql import SparkSession
from pyspark.pipelines.block_session_mutations import block_session_mutations
from pyspark.pipelines.graph_element_registry import (
    graph_element_registration_context,
    GraphElementRegistry,
)
from pyspark.pipelines.init_cli import init
from pyspark.pipelines.logging_utils import log_with_curr_timestamp
from pyspark.pipelines.spark_connect_graph_element_registry import (
    SparkConnectGraphElementRegistry,
)
from pyspark.pipelines.spark_connect_pipeline import (
    create_dataflow_graph,
    start_run,
    handle_pipeline_events,
)

from pyspark.pipelines.add_pipeline_analysis_context import add_pipeline_analysis_context

PIPELINE_SPEC_FILE_NAMES = ["spark-pipeline.yaml", "spark-pipeline.yml"]


@dataclass(frozen=True)
class LibrariesGlob:
    """A glob pattern for finding pipeline source codes."""

    include: str


def validate_patch_glob_pattern(glob_pattern: str) -> str:
    """Validates that a glob pattern is allowed.

    Only allows:
    - File paths (paths without wildcards except for the filename)
    - Folder paths ending with /** (recursive directory patterns)

    Disallows complex glob patterns like transformations/**/*.py
    """
    # Check if it's a simple file path (no wildcards at all)
    if not glob.has_magic(glob_pattern):
        return glob_pattern

    # Check if it's a folder path ending with /**
    if glob_pattern.endswith("/**"):
        prefix = glob_pattern[:-3]
        if not glob.has_magic(prefix):
            # append "/*" to match everything under the directory recursively
            return glob_pattern + "/*"

    raise PySparkException(
        errorClass="PIPELINE_SPEC_INVALID_GLOB_PATTERN",
        messageParameters={"glob_pattern": glob_pattern},
    )


@dataclass(frozen=True)
class PipelineSpec:
    """Spec for a pipeline.

    :param name: The name of the pipeline.
    :param storage: The root directory for storing metadata, such as streaming checkpoints.
    :param catalog: The default catalog to use for the pipeline.
    :param database: The default database to use for the pipeline.
    :param configuration: A dictionary of Spark configuration properties to set for the pipeline.
    :param libraries: A list of glob patterns for finding pipeline source codes.
    """

    name: str
    storage: str
    catalog: Optional[str]
    database: Optional[str]
    configuration: Mapping[str, str]
    libraries: Sequence[LibrariesGlob]

    def __post_init__(self) -> None:
        """Validate libraries automatically after instantiation."""
        validated = [
            LibrariesGlob(validate_patch_glob_pattern(lib.include)) for lib in self.libraries
        ]

        # If normalization changed anything, patch into frozen dataclass
        if tuple(validated) != tuple(self.libraries):
            object.__setattr__(self, "libraries", tuple(validated))


def find_pipeline_spec(current_dir: Path) -> Path:
    """Looks in the current directory and its ancestors for a pipeline spec file."""
    while True:
        try:
            candidates = [
                current_dir / spec_file_name for spec_file_name in PIPELINE_SPEC_FILE_NAMES
            ]
            found_files = [candidate for candidate in candidates if candidate.is_file()]
            if len(found_files) == 1:
                return found_files[0]
            elif len(found_files) > 1:
                raise PySparkException(
                    errorClass="MULTIPLE_PIPELINE_SPEC_FILES_FOUND",
                    messageParameters={"dir_path": str(current_dir)},
                )
        except PermissionError:
            raise PySparkException(
                errorClass="PIPELINE_SPEC_FILE_NOT_FOUND",
                messageParameters={"dir_path": str(current_dir)},
            )

        if current_dir.parent == current_dir or not current_dir.parent.exists():
            raise PySparkException(
                errorClass="PIPELINE_SPEC_FILE_NOT_FOUND",
                messageParameters={"dir_path": str(current_dir)},
            )

        current_dir = current_dir.parent


def load_pipeline_spec(spec_path: Path) -> PipelineSpec:
    """Load the pipeline spec from a YAML file at the given path."""
    with spec_path.open("r") as f:
        return unpack_pipeline_spec(yaml.safe_load(f))


def unpack_pipeline_spec(spec_data: Mapping[str, Any]) -> PipelineSpec:
    ALLOWED_FIELDS = {
        "name",
        "storage",
        "catalog",
        "database",
        "schema",
        "configuration",
        "libraries",
    }
    REQUIRED_FIELDS = ["name", "storage"]
    for key in spec_data.keys():
        if key not in ALLOWED_FIELDS:
            raise PySparkException(
                errorClass="PIPELINE_SPEC_UNEXPECTED_FIELD", messageParameters={"field_name": key}
            )

    for key in REQUIRED_FIELDS:
        if key not in spec_data:
            raise PySparkException(
                errorClass="PIPELINE_SPEC_MISSING_REQUIRED_FIELD",
                messageParameters={"field_name": key},
            )

    return PipelineSpec(
        name=spec_data["name"],
        storage=spec_data["storage"],
        catalog=spec_data.get("catalog"),
        database=spec_data.get("database", spec_data.get("schema")),
        configuration=validate_str_dict(spec_data.get("configuration", {}), "configuration"),
        libraries=[
            LibrariesGlob(include=entry["glob"]["include"])
            for entry in spec_data.get("libraries", [])
        ],
    )


def validate_str_dict(d: Mapping[str, str], field_name: str) -> Mapping[str, str]:
    """Raises an error if the dictionary is not a mapping of strings to strings."""
    if not isinstance(d, dict):
        raise PySparkTypeError(
            errorClass="PIPELINE_SPEC_FIELD_NOT_DICT",
            messageParameters={"field_name": field_name, "field_type": type(d).__name__},
        )

    for key, value in d.items():
        if not isinstance(key, str):
            raise PySparkTypeError(
                errorClass="PIPELINE_SPEC_DICT_KEY_NOT_STRING",
                messageParameters={"field_name": field_name, "key_type": type(key).__name__},
            )
        if not isinstance(value, str):
            raise PySparkTypeError(
                errorClass="PIPELINE_SPEC_DICT_VALUE_NOT_STRING",
                messageParameters={
                    "field_name": field_name,
                    "key_name": key,
                    "value_type": type(value).__name__,
                },
            )

    return d


def register_definitions(
    spec_path: Path,
    registry: GraphElementRegistry,
    spec: PipelineSpec,
    spark: SparkSession,
    dataflow_graph_id: str,
) -> None:
    """Register the graph element definitions in the pipeline spec with the given registry.
    - Import Python files matching the glob patterns in the spec.
    - Register SQL files matching the glob patterns in the spec.
    """
    path = spec_path.parent.resolve()

    with change_dir(path):
        with graph_element_registration_context(registry):
            log_with_curr_timestamp(f"Loading definitions. Root directory: '{path}'.")
            for libraries_glob in spec.libraries:
                glob_expression = libraries_glob.include
                matching_files = [
                    p
                    for p in path.glob(glob_expression)
                    if p.is_file() and "__pycache__" not in p.parts  # ignore generated python cache
                ]
                log_with_curr_timestamp(
                    f"Found {len(matching_files)} files matching glob '{glob_expression}'"
                )
                for file in matching_files:
                    if file.suffix == ".py":
                        log_with_curr_timestamp(f"Importing {file}...")
                        module_spec = importlib.util.spec_from_file_location(file.stem, str(file))
                        assert module_spec is not None, f"Could not find module spec for {file}"
                        module = importlib.util.module_from_spec(module_spec)
                        assert (
                            module_spec.loader is not None
                        ), f"Module spec has no loader for {file}"
                        with add_pipeline_analysis_context(
                            spark=spark, dataflow_graph_id=dataflow_graph_id, flow_name=None
                        ):
                            with block_session_mutations():
                                module_spec.loader.exec_module(module)
                    elif file.suffix == ".sql":
                        log_with_curr_timestamp(f"Registering SQL file {file}...")
                        with file.open("r") as f:
                            sql = f.read()
                        file_path_relative_to_spec = file.relative_to(path)
                        registry.register_sql(sql, file_path_relative_to_spec)
                    else:
                        raise PySparkException(
                            errorClass="PIPELINE_UNSUPPORTED_DEFINITIONS_FILE_EXTENSION",
                            messageParameters={"file_path": str(file)},
                        )


@contextmanager
def change_dir(path: Path) -> Generator[None, None, None]:
    """Change the current working directory to the given path and restore it on close()."""
    prev = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev)


def run(
    spec_path: Path,
    full_refresh: Sequence[str],
    full_refresh_all: bool,
    refresh: Sequence[str],
    dry: bool,
) -> None:
    """Run the pipeline defined with the given spec.

    :param spec_path: Path to the pipeline specification file.
    :param full_refresh: List of datasets to reset and recompute.
    :param full_refresh_all: Perform a full graph reset and recompute.
    :param refresh: List of datasets to update.
    """
    # Validate conflicting arguments
    if full_refresh_all:
        if full_refresh:
            raise PySparkException(
                errorClass="CONFLICTING_PIPELINE_REFRESH_OPTIONS",
                messageParameters={
                    "conflicting_option": "--full_refresh",
                },
            )
        if refresh:
            raise PySparkException(
                errorClass="CONFLICTING_PIPELINE_REFRESH_OPTIONS",
                messageParameters={
                    "conflicting_option": "--refresh",
                },
            )

    log_with_curr_timestamp(f"Loading pipeline spec from {spec_path}...")
    spec = load_pipeline_spec(spec_path)

    log_with_curr_timestamp("Creating Spark session...")
    spark_builder = SparkSession.builder.config(
        "spark.sql.connect.serverStacktrace.enabled", "false"
    )
    for key, value in spec.configuration.items():
        spark_builder = spark_builder.config(key, value)

    spark = spark_builder.getOrCreate()

    log_with_curr_timestamp("Creating dataflow graph...")
    dataflow_graph_id = create_dataflow_graph(
        spark,
        default_catalog=spec.catalog,
        default_database=spec.database,
        sql_conf=spec.configuration,
    )

    log_with_curr_timestamp("Registering graph elements...")
    registry = SparkConnectGraphElementRegistry(spark, dataflow_graph_id)
    register_definitions(spec_path, registry, spec, spark, dataflow_graph_id)

    log_with_curr_timestamp("Starting run...")
    result_iter = start_run(
        spark,
        dataflow_graph_id,
        full_refresh=full_refresh,
        full_refresh_all=full_refresh_all,
        refresh=refresh,
        dry=dry,
        storage=spec.storage,
    )
    try:
        handle_pipeline_events(result_iter)
    finally:
        spark.stop()


def parse_table_list(value: str) -> List[str]:
    """Parse a comma-separated list of table names, handling whitespace."""
    return [table.strip() for table in value.split(",") if table.strip()]


def main() -> None:
    """The entry point of spark-pipelines CLI."""
    parser = argparse.ArgumentParser(description="Pipelines CLI")
    subparsers = parser.add_subparsers(dest="command", required=True)

    # "run" subcommand
    run_parser = subparsers.add_parser(
        "run",
        help="Run a pipeline. If no refresh options specified, "
        "a default incremental update is performed.",
    )
    run_parser.add_argument("--spec", help="Path to the pipeline spec.")
    run_parser.add_argument(
        "--full-refresh",
        type=parse_table_list,
        action="extend",
        help="List of datasets to reset and recompute (comma-separated).",
        default=[],
    )
    run_parser.add_argument(
        "--full-refresh-all",
        action="store_true",
        help="Perform a full graph reset and recompute.",
    )
    run_parser.add_argument(
        "--refresh",
        type=parse_table_list,
        action="extend",
        help="List of datasets to update (comma-separated).",
        default=[],
    )

    # "dry-run" subcommand
    dry_run_parser = subparsers.add_parser(
        "dry-run",
        help="Launch a run that just validates the graph and checks for errors.",
    )
    dry_run_parser.add_argument("--spec", help="Path to the pipeline spec.")

    # "init" subcommand
    init_parser = subparsers.add_parser(
        "init",
        help="Generate a sample pipeline project, with a spec file and example transformations.",
    )
    init_parser.add_argument(
        "--name",
        help="Name of the project. A directory with this name will be created underneath the "
        "current directory.",
        required=True,
    )

    args = parser.parse_args()
    assert args.command in ["run", "dry-run", "init"]

    if args.command in ["run", "dry-run"]:
        if args.spec is not None:
            spec_path = Path(args.spec)
            if not spec_path.is_file():
                raise PySparkException(
                    errorClass="PIPELINE_SPEC_FILE_DOES_NOT_EXIST",
                    messageParameters={"spec_path": args.spec},
                )
        else:
            spec_path = find_pipeline_spec(Path.cwd())

        if args.command == "run":
            run(
                spec_path=spec_path,
                full_refresh=args.full_refresh,
                full_refresh_all=args.full_refresh_all,
                refresh=args.refresh,
                dry=False,
            )
        else:
            assert args.command == "dry-run"
            run(
                spec_path=spec_path,
                full_refresh=[],
                full_refresh_all=False,
                refresh=[],
                dry=True,
            )
    elif args.command == "init":
        init(args.name)


if __name__ == "__main__":
    main()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pyspark.errors import PySparkValueError, PySparkTypeError
from pyspark.sql.connect.utils import check_dependencies

check_dependencies(__name__)

from typing import Any, Dict, Optional, Union, cast
import warnings

from pyspark import _NoValue
from pyspark._globals import _NoValueType
from pyspark.sql.conf import RuntimeConfig as PySparkRuntimeConfig
from pyspark.sql.connect import proto
from pyspark.sql.connect.client import SparkConnectClient


class RuntimeConf:
    def __init__(self, client: SparkConnectClient) -> None:
        """Create a new RuntimeConfig."""
        self._client = client

    __init__.__doc__ = PySparkRuntimeConfig.__init__.__doc__

    def set(self, key: str, value: Union[str, int, bool]) -> None:
        if isinstance(value, bool):
            value = "true" if value else "false"
        elif isinstance(value, int):
            value = str(value)
        op_set = proto.ConfigRequest.Set(pairs=[proto.KeyValue(key=key, value=value)])
        operation = proto.ConfigRequest.Operation(set=op_set)
        result = self._client.config(operation)
        for warn in result.warnings:
            warnings.warn(warn)

    set.__doc__ = PySparkRuntimeConfig.set.__doc__

    def _set_all(self, configs: Dict[str, Union[str, int, bool]], silent: bool) -> None:
        conf_list = []
        for key, value in configs.items():
            if isinstance(value, bool):
                value = "true" if value else "false"
            elif isinstance(value, int):
                value = str(value)
            conf_list.append(proto.KeyValue(key=key, value=value))
        op_set = proto.ConfigRequest.Set(pairs=conf_list, silent=silent)
        operation = proto.ConfigRequest.Operation(set=op_set)
        result = self._client.config(operation)
        for warn in result.warnings:
            warnings.warn(warn)

    def get(
        self, key: str, default: Union[Optional[str], _NoValueType] = _NoValue
    ) -> Optional[str]:
        self._checkType(key, "key")
        if default is _NoValue:
            op_get = proto.ConfigRequest.Get(keys=[key])
            operation = proto.ConfigRequest.Operation(get=op_get)
        else:
            if default is not None:
                self._checkType(default, "default")
            op_get_with_default = proto.ConfigRequest.GetWithDefault(
                pairs=[proto.KeyValue(key=key, value=cast(Optional[str], default))]
            )
            operation = proto.ConfigRequest.Operation(get_with_default=op_get_with_default)
        result = self._client.config(operation)
        return result.pairs[0][1]

    get.__doc__ = PySparkRuntimeConfig.get.__doc__

    @property
    def getAll(self) -> Dict[str, str]:
        op_get_all = proto.ConfigRequest.GetAll()
        operation = proto.ConfigRequest.Operation(get_all=op_get_all)
        result = self._client.config(operation)
        confs: Dict[str, str] = dict()
        for key, value in result.pairs:
            assert value is not None
            confs[key] = value
        return confs

    getAll.__doc__ = PySparkRuntimeConfig.getAll.__doc__

    def unset(self, key: str) -> None:
        op_unset = proto.ConfigRequest.Unset(keys=[key])
        operation = proto.ConfigRequest.Operation(unset=op_unset)
        result = self._client.config(operation)
        for warn in result.warnings:
            warnings.warn(warn)

    unset.__doc__ = PySparkRuntimeConfig.unset.__doc__

    def isModifiable(self, key: str) -> bool:
        op_is_modifiable = proto.ConfigRequest.IsModifiable(keys=[key])
        operation = proto.ConfigRequest.Operation(is_modifiable=op_is_modifiable)
        result = self._client.config(operation).pairs[0][1]
        if result == "true":
            return True
        elif result == "false":
            return False
        else:
            raise PySparkValueError(
                errorClass="VALUE_NOT_ALLOWED",
                messageParameters={"arg_name": "result", "allowed_values": "'true' or 'false'"},
            )

    isModifiable.__doc__ = PySparkRuntimeConfig.isModifiable.__doc__

    def _checkType(self, obj: Any, identifier: str) -> None:
        """Assert that an object is of type str."""
        if not isinstance(obj, str):
            raise PySparkTypeError(
                errorClass="NOT_STR",
                messageParameters={
                    "arg_name": identifier,
                    "arg_type": type(obj).__name__,
                },
            )


RuntimeConf.__doc__ = PySparkRuntimeConfig.__doc__


def _test() -> None:
    import os
    import sys
    import doctest
    from pyspark.sql import SparkSession as PySparkSession
    import pyspark.sql.connect.conf

    globs = pyspark.sql.connect.conf.__dict__.copy()
    globs["spark"] = (
        PySparkSession.builder.appName("sql.connect.conf tests")
        .remote(os.environ.get("SPARK_CONNECT_TESTING_REMOTE", "local[4]"))
        .getOrCreate()
    )

    (failure_count, test_count) = doctest.testmod(
        pyspark.sql.connect.conf,
        globs=globs,
        optionflags=doctest.ELLIPSIS
        | doctest.NORMALIZE_WHITESPACE
        | doctest.IGNORE_EXCEPTION_DETAIL,
    )

    globs["spark"].stop()

    if failure_count:
        sys.exit(-1)


if __name__ == "__main__":
    _test()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file

Licensed to the Apache Software Foundation (ASF) under one or more
contributor license agreements.  See the NOTICE file distributed with
this work for additional information regarding copyright ownership.
The ASF licenses this file to You under the Apache License, Version 2.0
(the "License"); you may not use this file except in compliance with
the License.  You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _HandleState:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _HandleStateEnumTypeWrapper(
    google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_HandleState.ValueType],
    builtins.type,
):  # noqa: F821
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    PRE_INIT: _HandleState.ValueType  # 0
    CREATED: _HandleState.ValueType  # 1
    INITIALIZED: _HandleState.ValueType  # 2
    DATA_PROCESSED: _HandleState.ValueType  # 3
    TIMER_PROCESSED: _HandleState.ValueType  # 4
    CLOSED: _HandleState.ValueType  # 5

class HandleState(_HandleState, metaclass=_HandleStateEnumTypeWrapper): ...

PRE_INIT: HandleState.ValueType  # 0
CREATED: HandleState.ValueType  # 1
INITIALIZED: HandleState.ValueType  # 2
DATA_PROCESSED: HandleState.ValueType  # 3
TIMER_PROCESSED: HandleState.ValueType  # 4
CLOSED: HandleState.ValueType  # 5
global___HandleState = HandleState

class StateRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VERSION_FIELD_NUMBER: builtins.int
    STATEFULPROCESSORCALL_FIELD_NUMBER: builtins.int
    STATEVARIABLEREQUEST_FIELD_NUMBER: builtins.int
    IMPLICITGROUPINGKEYREQUEST_FIELD_NUMBER: builtins.int
    TIMERREQUEST_FIELD_NUMBER: builtins.int
    UTILSREQUEST_FIELD_NUMBER: builtins.int
    version: builtins.int
    @property
    def statefulProcessorCall(self) -> global___StatefulProcessorCall: ...
    @property
    def stateVariableRequest(self) -> global___StateVariableRequest: ...
    @property
    def implicitGroupingKeyRequest(self) -> global___ImplicitGroupingKeyRequest: ...
    @property
    def timerRequest(self) -> global___TimerRequest: ...
    @property
    def utilsRequest(self) -> global___UtilsRequest: ...
    def __init__(
        self,
        *,
        version: builtins.int = ...,
        statefulProcessorCall: global___StatefulProcessorCall | None = ...,
        stateVariableRequest: global___StateVariableRequest | None = ...,
        implicitGroupingKeyRequest: global___ImplicitGroupingKeyRequest | None = ...,
        timerRequest: global___TimerRequest | None = ...,
        utilsRequest: global___UtilsRequest | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "implicitGroupingKeyRequest",
            b"implicitGroupingKeyRequest",
            "method",
            b"method",
            "stateVariableRequest",
            b"stateVariableRequest",
            "statefulProcessorCall",
            b"statefulProcessorCall",
            "timerRequest",
            b"timerRequest",
            "utilsRequest",
            b"utilsRequest",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "implicitGroupingKeyRequest",
            b"implicitGroupingKeyRequest",
            "method",
            b"method",
            "stateVariableRequest",
            b"stateVariableRequest",
            "statefulProcessorCall",
            b"statefulProcessorCall",
            "timerRequest",
            b"timerRequest",
            "utilsRequest",
            b"utilsRequest",
            "version",
            b"version",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["method", b"method"]
    ) -> (
        typing_extensions.Literal[
            "statefulProcessorCall",
            "stateVariableRequest",
            "implicitGroupingKeyRequest",
            "timerRequest",
            "utilsRequest",
        ]
        | None
    ): ...

global___StateRequest = StateRequest

class StateResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATUSCODE_FIELD_NUMBER: builtins.int
    ERRORMESSAGE_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    statusCode: builtins.int
    errorMessage: builtins.str
    value: builtins.bytes
    def __init__(
        self,
        *,
        statusCode: builtins.int = ...,
        errorMessage: builtins.str = ...,
        value: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "errorMessage", b"errorMessage", "statusCode", b"statusCode", "value", b"value"
        ],
    ) -> None: ...

global___StateResponse = StateResponse

class StateResponseWithLongTypeVal(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATUSCODE_FIELD_NUMBER: builtins.int
    ERRORMESSAGE_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    statusCode: builtins.int
    errorMessage: builtins.str
    value: builtins.int
    def __init__(
        self,
        *,
        statusCode: builtins.int = ...,
        errorMessage: builtins.str = ...,
        value: builtins.int = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "errorMessage", b"errorMessage", "statusCode", b"statusCode", "value", b"value"
        ],
    ) -> None: ...

global___StateResponseWithLongTypeVal = StateResponseWithLongTypeVal

class StateResponseWithStringTypeVal(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATUSCODE_FIELD_NUMBER: builtins.int
    ERRORMESSAGE_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    statusCode: builtins.int
    errorMessage: builtins.str
    value: builtins.str
    def __init__(
        self,
        *,
        statusCode: builtins.int = ...,
        errorMessage: builtins.str = ...,
        value: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "errorMessage", b"errorMessage", "statusCode", b"statusCode", "value", b"value"
        ],
    ) -> None: ...

global___StateResponseWithStringTypeVal = StateResponseWithStringTypeVal

class StateResponseWithListGet(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATUSCODE_FIELD_NUMBER: builtins.int
    ERRORMESSAGE_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    REQUIRENEXTFETCH_FIELD_NUMBER: builtins.int
    statusCode: builtins.int
    errorMessage: builtins.str
    @property
    def value(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]: ...
    requireNextFetch: builtins.bool
    def __init__(
        self,
        *,
        statusCode: builtins.int = ...,
        errorMessage: builtins.str = ...,
        value: collections.abc.Iterable[builtins.bytes] | None = ...,
        requireNextFetch: builtins.bool = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "errorMessage",
            b"errorMessage",
            "requireNextFetch",
            b"requireNextFetch",
            "statusCode",
            b"statusCode",
            "value",
            b"value",
        ],
    ) -> None: ...

global___StateResponseWithListGet = StateResponseWithListGet

class StateResponseWithMapKeysOrValues(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATUSCODE_FIELD_NUMBER: builtins.int
    ERRORMESSAGE_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    REQUIRENEXTFETCH_FIELD_NUMBER: builtins.int
    statusCode: builtins.int
    errorMessage: builtins.str
    @property
    def value(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]: ...
    requireNextFetch: builtins.bool
    def __init__(
        self,
        *,
        statusCode: builtins.int = ...,
        errorMessage: builtins.str = ...,
        value: collections.abc.Iterable[builtins.bytes] | None = ...,
        requireNextFetch: builtins.bool = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "errorMessage",
            b"errorMessage",
            "requireNextFetch",
            b"requireNextFetch",
            "statusCode",
            b"statusCode",
            "value",
            b"value",
        ],
    ) -> None: ...

global___StateResponseWithMapKeysOrValues = StateResponseWithMapKeysOrValues

class KeyAndValuePair(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KEY_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    key: builtins.bytes
    value: builtins.bytes
    def __init__(
        self,
        *,
        key: builtins.bytes = ...,
        value: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]
    ) -> None: ...

global___KeyAndValuePair = KeyAndValuePair

class StateResponseWithMapIterator(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATUSCODE_FIELD_NUMBER: builtins.int
    ERRORMESSAGE_FIELD_NUMBER: builtins.int
    KVPAIR_FIELD_NUMBER: builtins.int
    REQUIRENEXTFETCH_FIELD_NUMBER: builtins.int
    statusCode: builtins.int
    errorMessage: builtins.str
    @property
    def kvPair(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        global___KeyAndValuePair
    ]: ...
    requireNextFetch: builtins.bool
    def __init__(
        self,
        *,
        statusCode: builtins.int = ...,
        errorMessage: builtins.str = ...,
        kvPair: collections.abc.Iterable[global___KeyAndValuePair] | None = ...,
        requireNextFetch: builtins.bool = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "errorMessage",
            b"errorMessage",
            "kvPair",
            b"kvPair",
            "requireNextFetch",
            b"requireNextFetch",
            "statusCode",
            b"statusCode",
        ],
    ) -> None: ...

global___StateResponseWithMapIterator = StateResponseWithMapIterator

class TimerInfo(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KEY_FIELD_NUMBER: builtins.int
    TIMESTAMPMS_FIELD_NUMBER: builtins.int
    key: builtins.bytes
    timestampMs: builtins.int
    def __init__(
        self,
        *,
        key: builtins.bytes | None = ...,
        timestampMs: builtins.int = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["_key", b"_key", "key", b"key"]
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "_key", b"_key", "key", b"key", "timestampMs", b"timestampMs"
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["_key", b"_key"]
    ) -> typing_extensions.Literal["key"] | None: ...

global___TimerInfo = TimerInfo

class StateResponseWithTimer(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATUSCODE_FIELD_NUMBER: builtins.int
    ERRORMESSAGE_FIELD_NUMBER: builtins.int
    TIMER_FIELD_NUMBER: builtins.int
    REQUIRENEXTFETCH_FIELD_NUMBER: builtins.int
    statusCode: builtins.int
    errorMessage: builtins.str
    @property
    def timer(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        global___TimerInfo
    ]: ...
    requireNextFetch: builtins.bool
    def __init__(
        self,
        *,
        statusCode: builtins.int = ...,
        errorMessage: builtins.str = ...,
        timer: collections.abc.Iterable[global___TimerInfo] | None = ...,
        requireNextFetch: builtins.bool = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "errorMessage",
            b"errorMessage",
            "requireNextFetch",
            b"requireNextFetch",
            "statusCode",
            b"statusCode",
            "timer",
            b"timer",
        ],
    ) -> None: ...

global___StateResponseWithTimer = StateResponseWithTimer

class StatefulProcessorCall(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SETHANDLESTATE_FIELD_NUMBER: builtins.int
    GETVALUESTATE_FIELD_NUMBER: builtins.int
    GETLISTSTATE_FIELD_NUMBER: builtins.int
    GETMAPSTATE_FIELD_NUMBER: builtins.int
    TIMERSTATECALL_FIELD_NUMBER: builtins.int
    DELETEIFEXISTS_FIELD_NUMBER: builtins.int
    @property
    def setHandleState(self) -> global___SetHandleState: ...
    @property
    def getValueState(self) -> global___StateCallCommand: ...
    @property
    def getListState(self) -> global___StateCallCommand: ...
    @property
    def getMapState(self) -> global___StateCallCommand: ...
    @property
    def timerStateCall(self) -> global___TimerStateCallCommand: ...
    @property
    def deleteIfExists(self) -> global___StateCallCommand: ...
    def __init__(
        self,
        *,
        setHandleState: global___SetHandleState | None = ...,
        getValueState: global___StateCallCommand | None = ...,
        getListState: global___StateCallCommand | None = ...,
        getMapState: global___StateCallCommand | None = ...,
        timerStateCall: global___TimerStateCallCommand | None = ...,
        deleteIfExists: global___StateCallCommand | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "deleteIfExists",
            b"deleteIfExists",
            "getListState",
            b"getListState",
            "getMapState",
            b"getMapState",
            "getValueState",
            b"getValueState",
            "method",
            b"method",
            "setHandleState",
            b"setHandleState",
            "timerStateCall",
            b"timerStateCall",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "deleteIfExists",
            b"deleteIfExists",
            "getListState",
            b"getListState",
            "getMapState",
            b"getMapState",
            "getValueState",
            b"getValueState",
            "method",
            b"method",
            "setHandleState",
            b"setHandleState",
            "timerStateCall",
            b"timerStateCall",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["method", b"method"]
    ) -> (
        typing_extensions.Literal[
            "setHandleState",
            "getValueState",
            "getListState",
            "getMapState",
            "timerStateCall",
            "deleteIfExists",
        ]
        | None
    ): ...

global___StatefulProcessorCall = StatefulProcessorCall

class StateVariableRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUESTATECALL_FIELD_NUMBER: builtins.int
    LISTSTATECALL_FIELD_NUMBER: builtins.int
    MAPSTATECALL_FIELD_NUMBER: builtins.int
    @property
    def valueStateCall(self) -> global___ValueStateCall: ...
    @property
    def listStateCall(self) -> global___ListStateCall: ...
    @property
    def mapStateCall(self) -> global___MapStateCall: ...
    def __init__(
        self,
        *,
        valueStateCall: global___ValueStateCall | None = ...,
        listStateCall: global___ListStateCall | None = ...,
        mapStateCall: global___MapStateCall | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "listStateCall",
            b"listStateCall",
            "mapStateCall",
            b"mapStateCall",
            "method",
            b"method",
            "valueStateCall",
            b"valueStateCall",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "listStateCall",
            b"listStateCall",
            "mapStateCall",
            b"mapStateCall",
            "method",
            b"method",
            "valueStateCall",
            b"valueStateCall",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["method", b"method"]
    ) -> typing_extensions.Literal["valueStateCall", "listStateCall", "mapStateCall"] | None: ...

global___StateVariableRequest = StateVariableRequest

class ImplicitGroupingKeyRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SETIMPLICITKEY_FIELD_NUMBER: builtins.int
    REMOVEIMPLICITKEY_FIELD_NUMBER: builtins.int
    @property
    def setImplicitKey(self) -> global___SetImplicitKey: ...
    @property
    def removeImplicitKey(self) -> global___RemoveImplicitKey: ...
    def __init__(
        self,
        *,
        setImplicitKey: global___SetImplicitKey | None = ...,
        removeImplicitKey: global___RemoveImplicitKey | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "method",
            b"method",
            "removeImplicitKey",
            b"removeImplicitKey",
            "setImplicitKey",
            b"setImplicitKey",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "method",
            b"method",
            "removeImplicitKey",
            b"removeImplicitKey",
            "setImplicitKey",
            b"setImplicitKey",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["method", b"method"]
    ) -> typing_extensions.Literal["setImplicitKey", "removeImplicitKey"] | None: ...

global___ImplicitGroupingKeyRequest = ImplicitGroupingKeyRequest

class TimerRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TIMERVALUEREQUEST_FIELD_NUMBER: builtins.int
    EXPIRYTIMERREQUEST_FIELD_NUMBER: builtins.int
    @property
    def timerValueRequest(self) -> global___TimerValueRequest: ...
    @property
    def expiryTimerRequest(self) -> global___ExpiryTimerRequest: ...
    def __init__(
        self,
        *,
        timerValueRequest: global___TimerValueRequest | None = ...,
        expiryTimerRequest: global___ExpiryTimerRequest | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "expiryTimerRequest",
            b"expiryTimerRequest",
            "method",
            b"method",
            "timerValueRequest",
            b"timerValueRequest",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "expiryTimerRequest",
            b"expiryTimerRequest",
            "method",
            b"method",
            "timerValueRequest",
            b"timerValueRequest",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["method", b"method"]
    ) -> typing_extensions.Literal["timerValueRequest", "expiryTimerRequest"] | None: ...

global___TimerRequest = TimerRequest

class TimerValueRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    GETPROCESSINGTIMER_FIELD_NUMBER: builtins.int
    GETWATERMARK_FIELD_NUMBER: builtins.int
    @property
    def getProcessingTimer(self) -> global___GetProcessingTime: ...
    @property
    def getWatermark(self) -> global___GetWatermark: ...
    def __init__(
        self,
        *,
        getProcessingTimer: global___GetProcessingTime | None = ...,
        getWatermark: global___GetWatermark | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "getProcessingTimer",
            b"getProcessingTimer",
            "getWatermark",
            b"getWatermark",
            "method",
            b"method",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "getProcessingTimer",
            b"getProcessingTimer",
            "getWatermark",
            b"getWatermark",
            "method",
            b"method",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["method", b"method"]
    ) -> typing_extensions.Literal["getProcessingTimer", "getWatermark"] | None: ...

global___TimerValueRequest = TimerValueRequest

class ExpiryTimerRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ITERATORID_FIELD_NUMBER: builtins.int
    EXPIRYTIMESTAMPMS_FIELD_NUMBER: builtins.int
    iteratorId: builtins.str
    expiryTimestampMs: builtins.int
    def __init__(
        self,
        *,
        iteratorId: builtins.str = ...,
        expiryTimestampMs: builtins.int = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "expiryTimestampMs", b"expiryTimestampMs", "iteratorId", b"iteratorId"
        ],
    ) -> None: ...

global___ExpiryTimerRequest = ExpiryTimerRequest

class GetProcessingTime(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___GetProcessingTime = GetProcessingTime

class GetWatermark(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___GetWatermark = GetWatermark

class UtilsRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PARSESTRINGSCHEMA_FIELD_NUMBER: builtins.int
    @property
    def parseStringSchema(self) -> global___ParseStringSchema: ...
    def __init__(
        self,
        *,
        parseStringSchema: global___ParseStringSchema | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "method", b"method", "parseStringSchema", b"parseStringSchema"
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "method", b"method", "parseStringSchema", b"parseStringSchema"
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["method", b"method"]
    ) -> typing_extensions.Literal["parseStringSchema"] | None: ...

global___UtilsRequest = UtilsRequest

class ParseStringSchema(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SCHEMA_FIELD_NUMBER: builtins.int
    schema: builtins.str
    def __init__(
        self,
        *,
        schema: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["schema", b"schema"]) -> None: ...

global___ParseStringSchema = ParseStringSchema

class StateCallCommand(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATENAME_FIELD_NUMBER: builtins.int
    SCHEMA_FIELD_NUMBER: builtins.int
    MAPSTATEVALUESCHEMA_FIELD_NUMBER: builtins.int
    TTL_FIELD_NUMBER: builtins.int
    stateName: builtins.str
    schema: builtins.str
    mapStateValueSchema: builtins.str
    @property
    def ttl(self) -> global___TTLConfig: ...
    def __init__(
        self,
        *,
        stateName: builtins.str = ...,
        schema: builtins.str = ...,
        mapStateValueSchema: builtins.str = ...,
        ttl: global___TTLConfig | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["ttl", b"ttl"]) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "mapStateValueSchema",
            b"mapStateValueSchema",
            "schema",
            b"schema",
            "stateName",
            b"stateName",
            "ttl",
            b"ttl",
        ],
    ) -> None: ...

global___StateCallCommand = StateCallCommand

class TimerStateCallCommand(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REGISTER_FIELD_NUMBER: builtins.int
    DELETE_FIELD_NUMBER: builtins.int
    LIST_FIELD_NUMBER: builtins.int
    @property
    def register(self) -> global___RegisterTimer: ...
    @property
    def delete(self) -> global___DeleteTimer: ...
    @property
    def list(self) -> global___ListTimers: ...
    def __init__(
        self,
        *,
        register: global___RegisterTimer | None = ...,
        delete: global___DeleteTimer | None = ...,
        list: global___ListTimers | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "delete", b"delete", "list", b"list", "method", b"method", "register", b"register"
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "delete", b"delete", "list", b"list", "method", b"method", "register", b"register"
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["method", b"method"]
    ) -> typing_extensions.Literal["register", "delete", "list"] | None: ...

global___TimerStateCallCommand = TimerStateCallCommand

class ValueStateCall(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATENAME_FIELD_NUMBER: builtins.int
    EXISTS_FIELD_NUMBER: builtins.int
    GET_FIELD_NUMBER: builtins.int
    VALUESTATEUPDATE_FIELD_NUMBER: builtins.int
    CLEAR_FIELD_NUMBER: builtins.int
    stateName: builtins.str
    @property
    def exists(self) -> global___Exists: ...
    @property
    def get(self) -> global___Get: ...
    @property
    def valueStateUpdate(self) -> global___ValueStateUpdate: ...
    @property
    def clear(self) -> global___Clear: ...
    def __init__(
        self,
        *,
        stateName: builtins.str = ...,
        exists: global___Exists | None = ...,
        get: global___Get | None = ...,
        valueStateUpdate: global___ValueStateUpdate | None = ...,
        clear: global___Clear | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "clear",
            b"clear",
            "exists",
            b"exists",
            "get",
            b"get",
            "method",
            b"method",
            "valueStateUpdate",
            b"valueStateUpdate",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "clear",
            b"clear",
            "exists",
            b"exists",
            "get",
            b"get",
            "method",
            b"method",
            "stateName",
            b"stateName",
            "valueStateUpdate",
            b"valueStateUpdate",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["method", b"method"]
    ) -> typing_extensions.Literal["exists", "get", "valueStateUpdate", "clear"] | None: ...

global___ValueStateCall = ValueStateCall

class ListStateCall(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATENAME_FIELD_NUMBER: builtins.int
    EXISTS_FIELD_NUMBER: builtins.int
    LISTSTATEGET_FIELD_NUMBER: builtins.int
    LISTSTATEPUT_FIELD_NUMBER: builtins.int
    APPENDVALUE_FIELD_NUMBER: builtins.int
    APPENDLIST_FIELD_NUMBER: builtins.int
    CLEAR_FIELD_NUMBER: builtins.int
    stateName: builtins.str
    @property
    def exists(self) -> global___Exists: ...
    @property
    def listStateGet(self) -> global___ListStateGet: ...
    @property
    def listStatePut(self) -> global___ListStatePut: ...
    @property
    def appendValue(self) -> global___AppendValue: ...
    @property
    def appendList(self) -> global___AppendList: ...
    @property
    def clear(self) -> global___Clear: ...
    def __init__(
        self,
        *,
        stateName: builtins.str = ...,
        exists: global___Exists | None = ...,
        listStateGet: global___ListStateGet | None = ...,
        listStatePut: global___ListStatePut | None = ...,
        appendValue: global___AppendValue | None = ...,
        appendList: global___AppendList | None = ...,
        clear: global___Clear | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "appendList",
            b"appendList",
            "appendValue",
            b"appendValue",
            "clear",
            b"clear",
            "exists",
            b"exists",
            "listStateGet",
            b"listStateGet",
            "listStatePut",
            b"listStatePut",
            "method",
            b"method",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "appendList",
            b"appendList",
            "appendValue",
            b"appendValue",
            "clear",
            b"clear",
            "exists",
            b"exists",
            "listStateGet",
            b"listStateGet",
            "listStatePut",
            b"listStatePut",
            "method",
            b"method",
            "stateName",
            b"stateName",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["method", b"method"]
    ) -> (
        typing_extensions.Literal[
            "exists", "listStateGet", "listStatePut", "appendValue", "appendList", "clear"
        ]
        | None
    ): ...

global___ListStateCall = ListStateCall

class MapStateCall(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATENAME_FIELD_NUMBER: builtins.int
    EXISTS_FIELD_NUMBER: builtins.int
    GETVALUE_FIELD_NUMBER: builtins.int
    CONTAINSKEY_FIELD_NUMBER: builtins.int
    UPDATEVALUE_FIELD_NUMBER: builtins.int
    ITERATOR_FIELD_NUMBER: builtins.int
    KEYS_FIELD_NUMBER: builtins.int
    VALUES_FIELD_NUMBER: builtins.int
    REMOVEKEY_FIELD_NUMBER: builtins.int
    CLEAR_FIELD_NUMBER: builtins.int
    stateName: builtins.str
    @property
    def exists(self) -> global___Exists: ...
    @property
    def getValue(self) -> global___GetValue: ...
    @property
    def containsKey(self) -> global___ContainsKey: ...
    @property
    def updateValue(self) -> global___UpdateValue: ...
    @property
    def iterator(self) -> global___Iterator: ...
    @property
    def keys(self) -> global___Keys: ...
    @property
    def values(self) -> global___Values: ...
    @property
    def removeKey(self) -> global___RemoveKey: ...
    @property
    def clear(self) -> global___Clear: ...
    def __init__(
        self,
        *,
        stateName: builtins.str = ...,
        exists: global___Exists | None = ...,
        getValue: global___GetValue | None = ...,
        containsKey: global___ContainsKey | None = ...,
        updateValue: global___UpdateValue | None = ...,
        iterator: global___Iterator | None = ...,
        keys: global___Keys | None = ...,
        values: global___Values | None = ...,
        removeKey: global___RemoveKey | None = ...,
        clear: global___Clear | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "clear",
            b"clear",
            "containsKey",
            b"containsKey",
            "exists",
            b"exists",
            "getValue",
            b"getValue",
            "iterator",
            b"iterator",
            "keys",
            b"keys",
            "method",
            b"method",
            "removeKey",
            b"removeKey",
            "updateValue",
            b"updateValue",
            "values",
            b"values",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "clear",
            b"clear",
            "containsKey",
            b"containsKey",
            "exists",
            b"exists",
            "getValue",
            b"getValue",
            "iterator",
            b"iterator",
            "keys",
            b"keys",
            "method",
            b"method",
            "removeKey",
            b"removeKey",
            "stateName",
            b"stateName",
            "updateValue",
            b"updateValue",
            "values",
            b"values",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["method", b"method"]
    ) -> (
        typing_extensions.Literal[
            "exists",
            "getValue",
            "containsKey",
            "updateValue",
            "iterator",
            "keys",
            "values",
            "removeKey",
            "clear",
        ]
        | None
    ): ...

global___MapStateCall = MapStateCall

class SetImplicitKey(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KEY_FIELD_NUMBER: builtins.int
    key: builtins.bytes
    def __init__(
        self,
        *,
        key: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["key", b"key"]) -> None: ...

global___SetImplicitKey = SetImplicitKey

class RemoveImplicitKey(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___RemoveImplicitKey = RemoveImplicitKey

class Exists(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___Exists = Exists

class Get(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___Get = Get

class RegisterTimer(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EXPIRYTIMESTAMPMS_FIELD_NUMBER: builtins.int
    expiryTimestampMs: builtins.int
    def __init__(
        self,
        *,
        expiryTimestampMs: builtins.int = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["expiryTimestampMs", b"expiryTimestampMs"]
    ) -> None: ...

global___RegisterTimer = RegisterTimer

class DeleteTimer(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EXPIRYTIMESTAMPMS_FIELD_NUMBER: builtins.int
    expiryTimestampMs: builtins.int
    def __init__(
        self,
        *,
        expiryTimestampMs: builtins.int = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["expiryTimestampMs", b"expiryTimestampMs"]
    ) -> None: ...

global___DeleteTimer = DeleteTimer

class ListTimers(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ITERATORID_FIELD_NUMBER: builtins.int
    iteratorId: builtins.str
    def __init__(
        self,
        *,
        iteratorId: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["iteratorId", b"iteratorId"]
    ) -> None: ...

global___ListTimers = ListTimers

class ValueStateUpdate(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUE_FIELD_NUMBER: builtins.int
    value: builtins.bytes
    def __init__(
        self,
        *,
        value: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["value", b"value"]) -> None: ...

global___ValueStateUpdate = ValueStateUpdate

class Clear(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___Clear = Clear

class ListStateGet(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ITERATORID_FIELD_NUMBER: builtins.int
    iteratorId: builtins.str
    def __init__(
        self,
        *,
        iteratorId: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["iteratorId", b"iteratorId"]
    ) -> None: ...

global___ListStateGet = ListStateGet

class ListStatePut(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUE_FIELD_NUMBER: builtins.int
    FETCHWITHARROW_FIELD_NUMBER: builtins.int
    @property
    def value(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]: ...
    fetchWithArrow: builtins.bool
    def __init__(
        self,
        *,
        value: collections.abc.Iterable[builtins.bytes] | None = ...,
        fetchWithArrow: builtins.bool = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "fetchWithArrow", b"fetchWithArrow", "value", b"value"
        ],
    ) -> None: ...

global___ListStatePut = ListStatePut

class AppendValue(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUE_FIELD_NUMBER: builtins.int
    value: builtins.bytes
    def __init__(
        self,
        *,
        value: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["value", b"value"]) -> None: ...

global___AppendValue = AppendValue

class AppendList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUE_FIELD_NUMBER: builtins.int
    FETCHWITHARROW_FIELD_NUMBER: builtins.int
    @property
    def value(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]: ...
    fetchWithArrow: builtins.bool
    def __init__(
        self,
        *,
        value: collections.abc.Iterable[builtins.bytes] | None = ...,
        fetchWithArrow: builtins.bool = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "fetchWithArrow", b"fetchWithArrow", "value", b"value"
        ],
    ) -> None: ...

global___AppendList = AppendList

class GetValue(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    USERKEY_FIELD_NUMBER: builtins.int
    userKey: builtins.bytes
    def __init__(
        self,
        *,
        userKey: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["userKey", b"userKey"]) -> None: ...

global___GetValue = GetValue

class ContainsKey(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    USERKEY_FIELD_NUMBER: builtins.int
    userKey: builtins.bytes
    def __init__(
        self,
        *,
        userKey: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["userKey", b"userKey"]) -> None: ...

global___ContainsKey = ContainsKey

class UpdateValue(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    USERKEY_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    userKey: builtins.bytes
    value: builtins.bytes
    def __init__(
        self,
        *,
        userKey: builtins.bytes = ...,
        value: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["userKey", b"userKey", "value", b"value"]
    ) -> None: ...

global___UpdateValue = UpdateValue

class Iterator(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ITERATORID_FIELD_NUMBER: builtins.int
    iteratorId: builtins.str
    def __init__(
        self,
        *,
        iteratorId: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["iteratorId", b"iteratorId"]
    ) -> None: ...

global___Iterator = Iterator

class Keys(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ITERATORID_FIELD_NUMBER: builtins.int
    iteratorId: builtins.str
    def __init__(
        self,
        *,
        iteratorId: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["iteratorId", b"iteratorId"]
    ) -> None: ...

global___Keys = Keys

class Values(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ITERATORID_FIELD_NUMBER: builtins.int
    iteratorId: builtins.str
    def __init__(
        self,
        *,
        iteratorId: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["iteratorId", b"iteratorId"]
    ) -> None: ...

global___Values = Values

class RemoveKey(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    USERKEY_FIELD_NUMBER: builtins.int
    userKey: builtins.bytes
    def __init__(
        self,
        *,
        userKey: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["userKey", b"userKey"]) -> None: ...

global___RemoveKey = RemoveKey

class SetHandleState(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATE_FIELD_NUMBER: builtins.int
    state: global___HandleState.ValueType
    def __init__(
        self,
        *,
        state: global___HandleState.ValueType = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["state", b"state"]) -> None: ...

global___SetHandleState = SetHandleState

class TTLConfig(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DURATIONMS_FIELD_NUMBER: builtins.int
    durationMs: builtins.int
    def __init__(
        self,
        *,
        durationMs: builtins.int = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["durationMs", b"durationMs"]
    ) -> None: ...

global___TTLConfig = TTLConfig
